module Fmt = Format

let is_ascii c = Char.code c = 10 || Char.code c > 30 && Char.code c < 128

let ascii_ratio buf =
  let k = ref 0 in
  Bytes.iter (fun c -> if is_ascii c then incr k) buf;
  float !k /. float (Bytes.length buf)

let dump ?(full = false) ff buf =
  if ascii_ratio buf > 0.80 then
    Fmt.fprintf ff "%s" (Bytes.to_string buf)
  else
  let buf =
    if not full && Bytes.length buf > 64 then
      Bytes.sub buf 0 64
    else
      buf
  in
  let i = ref 0 in
  Bytes.iter (fun c -> Fmt.fprintf ff "%02x" (Char.code c); incr i; if !i land 15 = 0 then Fmt.fprintf ff "\n") buf;
  if !i land 15 <> 0 then
    Fmt.fprintf ff "\n"

type config =
  { mutable self_addr : int64
  ; mutable self_id   : int64
  }

type hdr =
  { magic : int64
  ; id    : int64
  ; src   : int64
  ; dst   : int64
  ; flags : int32
  ; sz    : int32
  }

type job = Read | Write | Exec

type sockaddr_in =
  { fam  : int
  ; port : int
  ; ip   : int32
  }

let unpack_sockaddr_in buf = (Pack.unpack "sSi" buf : sockaddr_in)

let show_ip (ip : int32) =
  let buf = Pack.pack "i" (ref ip) in
  let a, b, c, d = (Pack.unpack "bbbb" buf : int * int * int * int) in
  Printf.sprintf "%d.%d.%d.%d" a b c d

let print_sin ff sin =
  Fmt.fprintf ff "{ fam = %d; port = %d; ip = %s }\n" sin.fam sin.port (show_ip sin.ip)

type peer =
  { addr  : int64
  ; sin   : sockaddr_in
  ; chan  : bytes
  }

let unpack_peer buf =
  { addr = Pack.unpack_int64 buf
  ; sin  = unpack_sockaddr_in (Bytes.sub buf 8 16)
  ; chan = Bytes.sub buf 24 0x218
  }

let print_peer ff p =
  Fmt.fprintf ff "%Lx " p.addr;
  print_sin ff p.sin;
  Fmt.fprintf ff "chan:\n";
  dump ~full:true ff p.chan

type t =
  | Ping of string
  | Pong of bytes
  | Unrec of bytes
  | Peering_request
  | Peering_reply of peer list
  | Transmit of bytes
  | End_transmit
  | Job of job * string

let unpack_hdr buf = (Pack.unpack "llllii" buf : hdr)

let pack_hdr hdr = Pack.pack "llllii" hdr

let print_hdr ff hdr =
  Fmt.fprintf ff "%Lx\n" hdr.magic;
  let id = Bytes.to_string (Pack.pack "l" (ref hdr.id)) in
  Fmt.fprintf ff "%s\n" id;
  Fmt.fprintf ff "%Lx -> %Lx\n" hdr.src hdr.dst;
  Fmt.fprintf ff "flags: %lx\n" hdr.flags;
  Fmt.fprintf ff "%ld bytes\n" hdr.sz


let print ff msg =
  match msg with
  | Ping s -> Fmt.fprintf ff "ping %s\n" s
  | Pong body -> Fmt.fprintf ff "pong:\n"; dump ff body
  | Peering_request -> Fmt.fprintf ff "peering request\n"
  | Peering_reply peers -> Fmt.fprintf ff "peering reply\n"; List.iter (print_peer ff) peers
  | Transmit body -> Fmt.fprintf ff "transmit:\n"; dump ff body
  | End_transmit -> Fmt.fprintf ff "end transmission\n"
  | Unrec body -> Fmt.fprintf ff "???\n"; dump ff body
  | Job (Exec, cmd) -> Fmt.fprintf ff "exec \"%s\"\n" cmd
  | Job (Write, cmd) -> Fmt.fprintf ff "write \"%s\"\n" cmd
  | Job (Read, cmd) -> Fmt.fprintf ff "read \"%s\"\n" cmd



let peering_request id src dst =
  let hdr =
    { magic = 0xd1d3c0de41414141L
    ; id
    ; src
    ; dst
    ; flags = 0x10000l
    ; sz    = 40l
    }
  in
  pack_hdr hdr


let ping id src dst ?(extra_sz = 0) msg =
  let hdr =
    { magic = 0xd1d3c0de41414141L
    ; id
    ; src
    ; dst
    ; flags = 0x100l
    ; sz    = Int32.of_int (40 + String.length msg + extra_sz)
    }
  in
  Bytes.cat (pack_hdr hdr) (Bytes.of_string msg)

let bad_magic =
  let hdr =
    { magic = 0xd1d3c0de42424242L
    ; id    = 0L
    ; src   = 0L
    ; dst   = 0L
    ; flags = 0l
    ; sz    = 40l
    }
  in
  pack_hdr hdr

let read cfg dst fname =
  let hdr =
    { magic = 0xd1d3c0de41414141L
    ; id    = cfg.self_id
    ; src   = cfg.self_addr
    ; dst
    ; flags = 0x204l
    ; sz    = Int32.of_int (40 + String.length fname + 1)
    }
  in
  Bytes.concat Bytes.empty [ (pack_hdr hdr); (Bytes.of_string fname); "\x00" ]


let parse_peering_reply buf =
  let rec loop buf acc =
    try
      let peer = unpack_peer (Bytes.sub buf 0 0x230) in
      loop (Bytes.sub buf 0x230 (Bytes.length buf - 0x230)) (peer :: acc)
    with _ ->
      List.rev acc
  in
  loop buf []

let parse buf =
  let hdr = unpack_hdr buf in
  assert (hdr.magic = 0xd1d3c0de41414141L);
  let body = Bytes.sub buf 40 (min (Bytes.length buf) (Int32.to_int hdr.sz) - 40) in
  let flags = Int32.to_int hdr.flags in
  let msg =
    if flags land 0x2000_000 <> 0 then
      Transmit body
    else
    if flags land 0x4000_000 <> 0 then
      End_transmit
    else
    if flags land 0x300 = 0x100 then
      if flags land 0x1000_000 = 0 then
        Ping (Bytes.to_string body)
      else
        Pong body
    else
    if flags land 0x300 = 0x200 then
      let s = Bytes.to_string body in
      match flags land 7 with
      | 1 -> Job (Exec, s)
      | 2 -> Job (Write, s)
      | 4 -> Job (Read, s)
      | _ -> Unrec body
    else
    if flags land 0x70000 = 0x10000 then
      if flags land 0x1000_000 = 0 then
        Peering_request
      else
        Peering_reply (parse_peering_reply body)
    else
      Unrec body
  in
  hdr, msg

